# pylint: disable=invalid-name
# pylint: disable=too-many-arguments
# pylint: disable=consider-using-f-string
# pylint: disable=too-many-locals

"""
Generate Foucault Knife Edge images (Foucaugrams) for mirror/lens testing.

Documentation and examples are available at <https://lenstest.readthedocs.io>
"""

import numpy as np
import matplotlib.pyplot as plt
import lenstest.lenstest

__all__ = ('gram',
           'plot_gram',
           'plot_knife_and_screen',
           'plot_lens_layout',
           'plot_mirror_layout')


def _transmitted(RoC, x_offset, z_offset, X, Y, conic=0, phi=0, mask=False, A=0):
    """
    Determine if X, Y points are transmitted past knife edge.

    This assumes that the point source of light is located at the center
    of the mirror madius of curvature, RoC (or two focal lengths from a lens).

    Args:
        RoC: radius of curvature of mirror [mm]
        x_offset: transverse knife edge displacement from optical axis [mm]
        z_offset: axial knife edge offset from paraxial focus [mm]
        X, Y: grid of points to evaluate [mm]
        conic: conic constant or Schwartzchild constant [-]
        phi: CCW rotation of knife edge from vertical [radians]
        mask: show knife edge without lens/mirror effects
        A: coefficient of spherical aberration
    Returns:
        1D boolean array describing points blocked by knife edge
    """
    if mask:
        sagitta = 0
    else:
        sagitta = lenstest.lenstest.sagitta(RoC, conic, X, Y, A=A)

    # rotate points
    Xr = X * np.cos(phi) + Y * np.sin(phi)

    # x values of rays intersecting plane of knife edge
    Lx = Xr * (-z_offset - sagitta * conic) / (RoC + sagitta * conic)

    # True/False array if points miss knife edge
    T_mask = Lx < x_offset

    return T_mask


def gram(D, RoC, x_offset, z_offset,
         conic=0, phi=0, N=100000, invert=False, on_grid=False, A=0, mask=False):
    """
    Create points that miss the knife edge.

    This assumes that the point source of light is located at the center
    of the mirror madius of curvature, RoC.

    The knife edge is displaced from the center of the mirror by x_offset.

    The knife edge is oriented its edge is vertical when phi=0.

    The conic section is specified by conic:
          conic = ∞ for surface that is flat
          conic > 0 for surface that is an oblate spheroid
          conic = 0 for surface that is a sphere
          0<conic<-1 for surface that is a prolate spheroid
          conic = -1 for surface that is a paraboloid
          conic < -1 for surface that is a hyperboloid

    Args:
        D: diameter of mirror [mm]
        RoC: radius of curvature of mirror [mm]
        x_offset: transverse knife edge displacement from optical axis [mm]
        z_offset: axial knife edge offset from paraxial focus [mm]
        conic: conic constant or Schwartzchild constant [-]
        phi: CCW rotation of knife edge from vertical [radians]
        N: number of points to generate
        invert: boolean to draw dark or light areas
        on_grid: if False generate points on a grid
        mask: show Ronchi ruling without lens/mirror effects
        A: spherical aberration [1/mm³]

    Returns:
        x, y: masked listed of points to plot
    """
    X, Y = lenstest.lenstest.XY_test_points(D, N=N, on_grid=on_grid)

    T_mask = _transmitted(RoC, x_offset, z_offset, X, Y,
                          conic=conic, phi=phi, A=A, mask=mask)

    if invert:
        T_mask = np.logical_not(T_mask)

    # hide all points that will appear dark
    x_image = np.ma.masked_where(T_mask, X)
    y_image = np.ma.masked_where(T_mask, Y)
    return x_image, y_image


def plot_gram(D, RoC, x_offset, z_offset,
              conic=0, phi=0, A=0, on_grid=False):
    """
    Plot the Foucault knife edge image.

    The idea is to graph both the beam on the knife edge and the expected
    projection on a screen located at the radius-of-curvature away from
    the focus.  This allows rapid visualization or roughly how much
    of the knife edge interacts with the screen.

    The beam size is limited by the diffraction focus limit of the beam
    (assuming a 1000nm wavelength).

    Args:
        D: diameter of mirror [mm]
        RoC: radius of curvature of mirror [mm]
        x_offset: transverse knife edge offset from optical axis [mm]
        z_offset: axial knife edge offset from paraxial focus [mm]
        conic: conic or Schwartzchild constant [-]
        phi: CCW rotation of Ronchi ruling from vertical [radians]
        init: set to False to allow updating plots
        on_grid: if False generate points on a grid
        invert: set to True to invert ruling

    Returns:
        fig: matplotlib Figure object representing the plot
        ax: matplotlib Axes object representing the plot
    """
    x, y = gram(D, RoC, x_offset, z_offset, conic=conic, phi=phi, A=A, on_grid=on_grid)

    lenstest.lenstest.draw_circle(D / 2, color='blue')
    plt.plot(x, y, 'o', markersize=0.1, color='white')
    plt.ylim(-D / 2 * 1.2, D / 2 * 1.2)
    plt.xlim(-D / 2 * 1.2, D / 2 * 1.2)
    plt.xlabel("(mm)")
    plt.ylabel("(mm)")
    plt.text(0.5, 0.97,
             "Mirror/Lens is %s " % lenstest.lenstest.conic_string(conic),
             color='white',
             ha='center', va='top',
             transform=plt.gca().transAxes
             )
    plt.title("Screen is %.0f mm from Focus" % RoC)
    plt.gca().set_aspect('equal')
    plt.gca().set_facecolor("black")


def plot_knife_and_screen(D, RoC, x_offset, z_offset,
                          conic=0, phi=0, init=True, A=0, on_grid=False):
    """
    Plot the Foucault knife edge image.

    The idea is to graph both the beam on the knife edge and the expected
    projection on a screen located at the radius-of-curvature away from
    the focus.  This allows rapid visualization or roughly how much
    of the knife edge interacts with the screen.

    The beam size is limited by the diffraction focus limit of the beam
    (assuming a 1000nm wavelength).

    Args:
        D: diameter of mirror [mm]
        RoC: radius of curvature of mirror [mm]
        x_offset: transverse knife edge offset from optical axis [mm]
        z_offset: axial knife edge offset from paraxial focus [mm]
        conic: conic or Schwartzchild constant [-]
        phi: CCW rotation of Ronchi ruling from vertical [radians]
        init: set to False to allow updating plots
        on_grid: if False generate points on a grid

    Returns:
        fig: matplotlib Figure object representing the plot
        ax: matplotlib Axes object representing the plot
    """

    if init:
        fig, ax = plt.subplots(1, 2, figsize=(10, 4))
    else:
        fig, ax = None, None

    # plot in the plane of the knife edge
    plt.subplot(1, 2, 1)
    plt.gca().set_aspect('equal')

    # draw the knife
    x, y = lenstest.lenstest.knife_polygon(D / 2, phi, x_offset)
    plt.fill(x, y, color='silver', alpha=0.8)
    plt.plot(0, 0, 'b+', markersize=5)

    # Draw circle showing spotsize in the knife edge plane
    r_geometric = abs(D / 2 / RoC * z_offset)        # mm
    r_diffraction = 0.5 * 1e-6 * (RoC / 2) / D  # mm
    r_spot = max(r_geometric, r_diffraction)
    lenstest.lenstest.draw_circle(r_spot, color='blue')

    # we always want to see the edge of the knife
    size = max(r_spot * 2.2, abs(2 * x_offset))
    plt.ylim(-size, size)
    plt.xlim(-size, size)

#    plt.text(0.5, 0.96,
#             'Δx=%.2fmm, ϕ=%.0f°' % (x_offset, np.degrees(phi)),
#             color='white',
#             ha='center', va='top',
#             transform=plt.gca().transAxes
#    )

    # label the plot
    plt.xlabel("Knife x (mm)")
    plt.ylabel("Knife y (mm)")
    ks = 'at'
    if z_offset < 0:
        ks = '%.3fmm before' % abs(z_offset)
    if z_offset > 0:
        ks = '%.3fmm after' % z_offset
    plt.title("%s focus" % ks)

    # plot in the plane of the screen
    plt.subplot(1, 2, 2)
    plt.gca().set_facecolor("black")
    plt.gca().set_aspect('equal')

    # knife edge result
    x, y = gram(D, RoC, x_offset, z_offset, conic=conic, phi=phi, A=A, on_grid=on_grid)
    plt.plot(x, y, 'o', markersize=0.1, color='white')

    # the circle for the projected circle from mirror
    lenstest.lenstest.draw_circle(D / 2, color='blue')
    size = D / 2 * 1.2
    plt.ylim(-size, size)
    plt.xlim(-size, size)

    # add labels
    plt.title("D=%.1fmm, RoC=%.1fmm, K=%.2f" % (D, RoC, conic))
    plt.xlabel("Screen x (mm)")
    plt.ylabel("Screen y (mm)")
    cs = lenstest.lenstest.conic_string(conic)
    plt.text(0.5, 0.97,
             "Mirror/Lens is %s " % cs,
             color='white',
             ha='center', va='top',
             transform=plt.gca().transAxes
             )
    plt.title("%.0fmm from Focus" % RoC)

    return fig, ax


def plot_lens_layout(D, RoC, x_offset, z_offset):
    """
    Plot the Foucault knife edge experiment.

    Args:
        D: diameter of mirror or lens [mm]
        RoC: radius of curvature of mirror [mm]
        x_offset: transverse knife edge offset from optical axis [mm]
        z_offset: axial knife edge offset from paraxial focus [mm]
    Returns:
        fig: matplotlib Figure object representing the plot
        ax: matplotlib Axes object representing the plot
    """
    fig, ax = plt.subplots(figsize=(10, 5))

    # slope of marginal ray and ray touching knife
    m2 = (D / 2) / RoC
    if z_offset == 0:
        m = 0
    else:
        m = x_offset / z_offset

    # account for knife completely missing or blocking beam
    m = min(m, m2)
    m = max(m, -m2)

    # marginal rays
    plt.plot([-2 * RoC, -RoC, RoC], [0, D / 2, -D / 2], color='black', linewidth=1)
    plt.plot([-2 * RoC, -RoC, RoC], [0, -D / 2, D / 2], color='black', linewidth=1)
    plt.plot(-2 * RoC, 0, 'ok', markersize=5)

    # ray touching the knife edge
    plt.plot([-2 * RoC, -RoC, RoC], [0, -m * RoC, m * RoC],
             color='black', linewidth=1, linestyle='--')

    # shade blocked light
    if z_offset < 0:
        plt.fill_between([z_offset, RoC], [m * z_offset, m * RoC],
                         [z_offset * m2, D / 2], color='darkgray')
    else:
        plt.fill_between([z_offset, RoC], [m * z_offset, m * RoC],
                         [-z_offset * m2, -D / 2], color='darkgray')

    # draw the lens
    lenstest.lenstest.draw_lens(D, RoC, -RoC)
    plt.text(-RoC, -D / 2, 'lens / mirror', ha='right', rotation=90, color='blue', clip_on=True)

# focus plane
#    plt.text(0, D / 2, ' focus', ha='left')
#    plt.axvline(0, color='black', linewidth = 1)

    # optical axis
    plt.axhline(0, color='blue', linewidth=1)
    plt.text(-RoC * 0.96, 0, 'optical axis ', ha='left', va='center', color='blue',
             bbox={"facecolor": "white", "edgecolor": "white"}, clip_on=True)

    # knife
#    plt.axvline(z_offset, color='black', linewidth = 0.5)
    plt.plot([z_offset, z_offset], [x_offset, -D / 2], lw=2, color='black')
    plt.text(z_offset, -D / 2, 'knife', ha='center', rotation=90, color='black',
             bbox={"facecolor": "white", "edgecolor": "white"}, clip_on=True)

    # screen
    plt.axvline(RoC, color='blue', linewidth=2)
    plt.text(RoC, -D / 2, 'projection screen', ha='left', rotation=90, color='blue', clip_on=True)

    plt.xlabel('Distance from focus (mm)')
    plt.ylabel('Height above optical axis (mm)')
    plt.title('Foucault Knife Test Showing Shading of Screen')
    return fig, ax


def plot_mirror_layout(D, RoC, x_offset, z_offset):
    """
    Plot the Foucault knife edge experiment.

    Args:
        D: diameter of mirror or lens [mm]
        RoC: radius of curvature of mirror [mm]
        x_offset: transverse offset of knife edge from optical axis [mm]
        z_offset: axial knife edge offset from paraxial focus [mm]
    Returns:
        fig: matplotlib Figure object representing the plot
        ax: matplotlib Axes object representing the plot
    """
    # slope of marginal ray and ray touching knife
    fig, ax = plt.subplots(figsize=(10, 5))

    m2 = (D / 2) / RoC
    if z_offset == 0:
        m = 0
    else:
        m = x_offset / z_offset

    # account for knife completely missing or blocking beam
    m = min(m, m2)
    m = max(m, -m2)

    # marginal rays
    plt.plot(0, D / 20, 'ok', markersize=5)
    plt.plot([0, -RoC, RoC], [D / 20, D / 2, -D / 2], color='black', linewidth=1)
    plt.plot([0, -RoC, RoC], [D / 20, -D / 2, D / 2], color='black', linewidth=1)

    # ray touching the knife edge
    plt.plot([0, -RoC, RoC], [D / 20, -m * RoC, m * RoC],
             color='black', linewidth=1, linestyle='--')

    # shade blocked light
    if z_offset < 0:
        plt.fill_between([z_offset, RoC], [m * z_offset, m * RoC],
                         [z_offset * m2, D / 2], color='darkgray')
    else:
        plt.fill_between([z_offset, RoC], [m * z_offset, m * RoC],
                         [-z_offset * m2, -D / 2], color='darkgray')

    # draw the lens
    lenstest.lenstest.draw_mirror(D, RoC, vertex=-RoC)
    plt.text(-RoC, -D / 2, 'lens / mirror', ha='right', rotation=90, color='blue', clip_on=True)

# focus plane
#    plt.text(0, D / 2, ' focus', ha='left')
#    plt.axvline(0, color='black', linewidth = 1)

    # optical axis
    plt.axhline(0, color='blue', linewidth=1)
    plt.text(-RoC * 0.96, 0, 'optical axis ', ha='left', va='center', color='blue',
             bbox={"facecolor": "white", "edgecolor": "white"}, clip_on=True)

    # knife
#    plt.axvline(z_offset, color='black', linewidth = 0.5)
    plt.plot([z_offset, z_offset], [x_offset, -D / 2], lw=2, color='black')
    plt.text(z_offset, -D / 2, 'knife', ha='center', rotation=90, color='black',
             bbox={"facecolor": "white", "edgecolor": "white"}, clip_on=True)

    # screen
    plt.axvline(RoC, color='blue', linewidth=2)
    plt.text(RoC, -D / 2, 'projection screen', ha='left', rotation=90, color='blue', clip_on=True)

    plt.xlabel('Distance from focus (mm)')
    plt.ylabel('Height above optical axis (mm)')
    plt.title('Foucault Knife Test Showing Shading of Screen')
    return fig, ax



# pylint: disable=invalid-name
# pylint: disable=protected-access
"""Test functionality of utility routines found in ronchi.py."""

import unittest
import numpy as np
import matplotlib.pyplot as plt
import lenstest


class TestGram(unittest.TestCase):
    """Evaluate gram functionality."""

    def test_gram_output_type(self):
        """Test that gram returns two numpy arrays."""
        D = 100
        RoC = 200
        lpm = 10
        z_offset = 10
        x, y = lenstest.ronchi.gram(D, RoC, lpm, z_offset)
        self.assertIsInstance(x, np.ndarray)
        self.assertIsInstance(y, np.ndarray)

    def test_gram_output_shape(self):
        """Test that gram returns arrays with correct shapes."""
        D = 100
        RoC = 200
        lpm = 10
        z_offset = 10
        x, y = lenstest.ronchi.gram(D, RoC, lpm, z_offset)
        self.assertEqual(x.shape, (100000,))
        self.assertEqual(y.shape, (100000,))

    def test_gram_invert_parameter(self):
        """Test that invert parameter inverts the output."""
        D = 100
        RoC = 200
        lpm = 10
        z_offset = 10
        x, y = lenstest.ronchi.gram(D, RoC, lpm, z_offset, invert=True)
        self.assertEqual(x.shape, (100000,))
        self.assertEqual(y.shape, (100000,))


class TestTransmitted(unittest.TestCase):
    """Evaluate _transmitted functionality."""

    def test_transmitted_output_type(self):
        """Test that _transmitted returns a numpy array."""
        RoC = 200
        lpm = 10
        z_offset = 10
        X = np.zeros((10, 10))
        Y = np.zeros((10, 10))
        result = lenstest.ronchi._transmitted(RoC, lpm, z_offset, X, Y)
        self.assertIsInstance(result, np.ndarray)

    def test_transmitted_output_shape(self):
        """Test that _transmitted returns array with correct shape."""
        RoC = 200
        lpm = 10
        z_offset = 10
        X = np.zeros((10, 10))
        Y = np.zeros((10, 10))
        result = lenstest.ronchi._transmitted(RoC, lpm, z_offset, X, Y)
        self.assertEqual(result.shape, (10, 10))

    def test_transmitted_mask_parameter(self):
        """Test that mask parameter works."""
        RoC = 200
        lpm = 10
        z_offset = 10
        X = np.zeros((10, 10))
        Y = np.zeros((10, 10))
        result = lenstest.ronchi._transmitted(RoC, lpm, z_offset, X, Y, mask=True)
        self.assertEqual(result.shape, (10, 10))


class TestPlotRulingAndScreen(unittest.TestCase):
    """Evaluate plot_ruling_and_screen functionality."""

    def test_plot_ruling_and_screen_output_type(self):
        """Test that plot_ruling_and_screen returns a tuple."""
        D = 100
        RoC = 200
        lpm = 10
        z_offset = 10
        fig, ax = lenstest.ronchi.plot_ruling_and_screen(D, RoC, lpm, z_offset)
        self.assertIsInstance(fig, plt.Figure)
        self.assertIsInstance(ax[0], plt.Axes)
        self.assertIsInstance(ax[1], plt.Axes)

    def test_plot_ruling_and_screen_output_shape(self):
        """Test that plot_ruling_and_screen returns a tuple with two elements."""
        D = 100
        RoC = 200
        lpm = 10
        z_offset = 10
        result = lenstest.ronchi.plot_ruling_and_screen(D, RoC, lpm, z_offset)
        self.assertEqual(len(result), 2)

    def test_plot_ruling_and_screen_grid_parameter(self):
        """Test that grid parameter has correct effect on plot."""
        D = 100
        RoC = 200
        lpm = 10
        z_offset = 10
        lenstest.ronchi.plot_ruling_and_screen(D, RoC, lpm, z_offset, on_grid=True)


class TestPlotLensLayout(unittest.TestCase):
    """Evaluate plot_lens_layout functionality."""

    def test_plot_lens_layout_output_type(self):
        """Test that plot_lens_layout returns a tuple of matplotlib Figure and Axes objects."""
        D = 100
        f = 200
        z_offset = 10
        fig, ax = lenstest.ronchi.plot_lens_layout(D, f, z_offset)
        self.assertIsInstance(fig, plt.Figure)
        self.assertIsInstance(ax, plt.Axes)

    def test_plot_lens_layout_output_shape(self):
        """Test that plot_lens_layout returns a tuple with two elements."""
        D = 100
        f = 200
        z_offset = 10
        result = lenstest.ronchi.plot_lens_layout(D, f, z_offset)
        self.assertEqual(len(result), 2)


class TestPlotMirrorLayout(unittest.TestCase):
    """Evaluate plot_mirror_layout functionality."""

    def test_plot_mirror_layout_output_type(self):
        """Test that plot_mirror_layout returns a tuple of matplotlib Figure and Axes objects."""
        D = 100
        RoC = 200
        z_offset = 10
        fig, ax = lenstest.ronchi.plot_lens_layout(D, RoC, z_offset)
        self.assertIsInstance(fig, plt.Figure)
        self.assertIsInstance(ax, plt.Axes)

    def test_plot_lens_layout_output_shape(self):
        """Test that plot_mirror_layout returns a tuple with two elements."""
        D = 100
        RoC = 200
        z_offset = 10
        result = lenstest.ronchi.plot_lens_layout(D, RoC, z_offset)
        self.assertEqual(len(result), 2)


# Run the test case
if __name__ == '__main__':
    unittest.main()



"""
Non-interferometric Lens Testing.

Documentation and examples are at <https://lenstest.readthedocs.io>

"""
__version__ = '0.9.0'
__author__ = 'Scott Prahl'
__email__ = 'scott.prahl@oit.edu'
__copyright__ = '2021-23, Scott Prahl'
__license__ = 'MIT'
__url__ = 'https://github.com/scottprahl/lenstest'

from .lenstest import *
from . import ronchi
from . import foucault



# pylint: disable=invalid-name
# pylint: disable=too-many-arguments
# pylint: disable=consider-using-f-string
# pylint: disable=too-many-locals
"""
Generate ronchigrams for lens testing.

Documentation and examples are available at <https://lenstest.readthedocs.io>
"""

import numpy as np
import matplotlib.pyplot as plt
import lenstest

__all__ = ('gram',
           'plot_gram',
           'plot_ruling_and_screen',
           'plot_lens_layout',
           'plot_mirror_layout')


def _transmitted(RoC, lpm, z_offset, X, Y, conic=0, mask=False, phi=0):
    """
    Determine if X, Y points are transmitted through Ronchi ruling.

    This assumes that the one of the opaque lines of the ruling is centered on
    the optical axis.

    This also assumes that the point source of light is located at the center
    of the mirror's madius of curvature.

    The lines of the ruling are vertical when phi=0.

    Finally, the ruling is located at RoC + z_offset.  Negative values indicate
    that the ruling is located in front of the focus of the mirror.

    If `mask==True` then the mirror parameters are ignored.  This is useful
    for creating a plot for light at the Ronchi ruling.

    Args:
        RoC: radius of curvature of mirror [mm]
        lpm: line pairs per mm [1/mm]
        z_offset: axial z_offset of grating from center of mirror's RoC [mm]
        X, Y: grid of points to evaluate [mm]
        conic: conic constant or Schwartzchild constant [-]
        mask: show Ronchi ruling without lens/mirror effects
        phi: CCW rotation of Ronchi ruling from horizontal [radians]
    Returns:
        1D boolean array describing points blocked by Ronchi ruling
    """
    if mask:
        sagitta = 0
    else:
        sagitta = lenstest.lenstest.sagitta(RoC, conic, X, Y)

    # rotate points
    Xr = X * np.cos(phi) + Y * np.sin(phi)

    # x values of rays intersecting Ronchi ruling plane
    Lx = Xr * (-z_offset - sagitta * conic) / (RoC + sagitta * conic)

    # scale so even values pass through ruling
    T = (np.abs(4 * lpm * Lx) + 0.5).astype(int)

    # True/False array that designates if points pass through ruling
    T_mask = T % 2 == 0

    return T_mask


def gram(D, RoC, lpm, z_offset, conic=0, phi=0,
         N=100000, invert=False, on_grid=False, mask=False):
    """
    Create points that pass through a Ronchi ruling.

    When plotted these points will be a Ronchigram.

    This assumes that the point source of light is located at the center
    of the mirror madius of curvature, RoC.

    The conic section is specified by conic:
          conic = ∞ for surface that is flat
          conic > 0 for surface that is an oblate spheroid
          conic = 0 for surface that is a sphere
          0<conic<-1 for surface that is a prolate spheroid
          conic = -1 for surface that is a paraboloid
          conic < -1 for surface that is a hyperboloid

    Args:
        D: diameter of mirror [mm]
        RoC: radius of curvature of mirror [mm]
        lpm: line pairs per mm [1/mm]
        z_offset: axial z_offset of grating from true focus [mm]
        conic: conic constant or Schwartzchild constant [-]
        phi: CCW rotation of Ronchi ruling from vertical
        N: number of points to generate
        invert: boolean to draw dark or light areas
        on_grid: if False generate points on a grid
        mask: show Ronchi ruling without lens/mirror effects

    Returns:
        x, y: masked listed of points to plot
    """
    X, Y = lenstest.lenstest.XY_test_points(D, N=N, on_grid=on_grid)

    T_mask = _transmitted(RoC, lpm, z_offset, X, Y, conic=conic, mask=mask, phi=phi)

    if invert:
        T_mask = np.logical_not(T_mask)

    # hide all points that should be dark
    x_mask = np.ma.masked_where(T_mask, X)
    y_mask = np.ma.masked_where(T_mask, Y)

    if mask:
        x_mask *= abs(z_offset / RoC)
        y_mask *= abs(z_offset / RoC)

    return x_mask, y_mask


def plot_gram(D, RoC, lpm, z_offset, conic=0, phi=0, on_grid=False, invert=False):
    """
    Plot cross-sections on projection screen at a distance of RoC.

    Args:
        D: diameter of mirror [mm]
        RoC: radius of curvature of mirror [mm]
        lpm: line pairs per mm [1/mm]
        conic: conic or Schwartzchild constant [-]
        phi: CCW rotation of Ronchi ruling from vertical [radians]
        on_grid: if False generate points on a grid
        invert: set to True to invert ruling

    Returns:
        Nothing.
    """
    # generate and plot all the points
    x, y = gram(D, RoC, lpm, z_offset, conic=conic, phi=phi, invert=invert, on_grid=on_grid)
    plt.plot(x, y, 'o', markersize=0.1, color='white')

    # Draw circle showing spotsize on projection screen
    lenstest.lenstest.draw_circle(D / 2, color='green')

    # limit plot to slightly larger than the beam size
    size = D / 2 * 1.2
    plt.ylim(-size, size)
    plt.xlim(-size, size)
    plt.gca().set_facecolor("black")
    plt.gca().set_aspect('equal')
    plt.xlabel("(mm)")
    plt.ylabel("(mm)")


def plot_ruling_and_screen(D, RoC, lpm, z_offset,
                           conic=0, phi=0, init=True, on_grid=False, invert=False):
    """
    Plot cross-sections at Ronchi ruling and projection screen.

    The idea is to graph both the beam on the ruling and the expected
    projection on a screen located at the radius-of-curvature away from
    the focus.  This allows rapid visualization or roughly how much
    of the Ronchi ruling interacts with the screen.

    The beam size is limited by the diffraction focus limit of the beam
    (assuming a 1000nm wavelength).

    Args:
        D: diameter of mirror [mm]
        RoC: radius of curvature of mirror [mm]
        lpm: line pairs per mm [1/mm]
        z_offset: axial z_offset of grating from true focus [mm]
        conic: conic or Schwartzchild constant [-]
        phi: CCW rotation of Ronchi ruling from vertical [radians]
        init: set to False to allow updating plots
        on_grid: if False generate points on a grid
        invert: set to True to invert ruling

    Returns:
        fig: matplotlib Figure object representing the plot
        ax: matplotlib Axes object representing the plot
    """
    if init:
        fig, ax = plt.subplots(1, 2, figsize=(10, 5))
    else:
        fig, ax = None, None

    # create plot in plane of Ronchi Ruling with beam size
    plt.subplot(1, 2, 1)
    plt.gca().set_facecolor("black")
    plt.gca().set_aspect('equal')

    # generate and plot all the points
    x, y = gram(D, RoC, lpm, z_offset, conic=conic,
                phi=phi, on_grid=on_grid, mask=True, invert=invert)
    plt.plot(x, y, 'o', markersize=0.6, color='white')

    # Draw circle showing spotsize at location of ruling
    r_geometric = abs(D / 2 / RoC * z_offset)        # mm
    r_diffraction = 0.5 * 1e-6 * (RoC / 2) / D       # mm
    r_spot = max(r_geometric, r_diffraction)
    lenstest.lenstest.draw_circle(r_spot, color='blue')
    if invert:
        plt.plot(0, 0, 'k+', markersize=5)
    else:
        plt.plot(0, 0, 'w+', markersize=5)

    # limit plot to slightly larger than the beam size
    size = r_spot * 1.2
    plt.ylim(-size, size)
    plt.xlim(-size, size)
    ks = 'at'
    if z_offset < 0:
        ks = '%.3fmm before' % abs(z_offset)
    if z_offset > 0:
        ks = '%.3fmm after' % z_offset
    plt.title("%s focus" % ks)
    plt.xlabel("Ruling Plane x (mm)")
    plt.ylabel("Ruling Plane y (mm)")

    # create plot in plane of the projection screen
    plt.subplot(1, 2, 2)
    plt.gca().set_facecolor("black")
    plt.gca().set_aspect('equal')

    # generate and plot all the points
    x, y = gram(D, RoC, lpm, z_offset,
                conic=conic, phi=phi, on_grid=on_grid, mask=False, invert=invert)
    plt.plot(x, y, 'o', markersize=0.1, color='white')

    # Draw circle showing spotsize on projection screen
    lenstest.lenstest.draw_circle(D / 2, color='green')

    # limit plot to slightly larger than the beam size
    size = D / 2 * 1.2
    plt.ylim(-size, size)
    plt.xlim(-size, size)
    plt.title('%.0fmm from Focus)' % RoC)
    plt.xlabel("Screen x (mm)")
    plt.ylabel("Screen y (mm)")

    return fig, ax


def plot_lens_layout(D, f, z_offset):
    """
    Plot the Ronchi Lens Test Layout (4f system).

    Args:
        D: diameter of mirror or lens [mm]
        f: focal length of lens [mm]
        z_offset: axial offset of knife edge from true focus [mm]
    Returns:
        fig: matplotlib Figure object representing the plot
        ax: matplotlib Axes object representing the plot
    """
    fig, ax = plt.subplots(figsize=(10, 5))

    RoC = f          # for lens assuming n=1.5 and biconvex lens

    # point source
    plt.text(-4 * f, D * 0.05, 'point source', ha='left', rotation=90, color='blue')

    # marginal rays with lens at -2f
    DD = 0.8 * D
    plt.plot([-4 * f, -2 * f, 2 * f], [0, DD / 2, -DD / 2], color='black', linewidth=1)
    plt.plot([-4 * f, -2 * f, 2 * f], [0, -DD / 2, DD / 2], color='black', linewidth=1)
    plt.plot([-4 * f, -2 * f, 2 * f], [0, DD / 4, -DD / 4], color='black', linewidth=1)
    plt.plot([-4 * f, -2 * f, 2 * f], [0, -DD / 4, DD / 4], color='black', linewidth=1)

    # draw the lens
    lenstest.lenstest.draw_lens(D, RoC, -2 * f)
    plt.text(-2 * f, D / 2, 'lens under test', ha='left', color='blue')

# focus plane
#    plt.text(0, D / 2, ' focus', ha='left')
#    plt.axvline(0, color='black', linewidth=1)

    # optical axis
    plt.axhline(0, color='blue', linewidth=1)
#    plt.text(-RoC * 0.9, 0, 'optical axis ', ha='left', va='center', color='blue',
#             bbox={"facecolor": "white", "edgecolor":"white"})

    # Ronchi
#    plt.axvline(z_offset, color='black', linewidth = 0.5)
    plt.plot([z_offset, z_offset], [D / 2, -D / 2], ls='--', lw=2, color='black')
    plt.text(z_offset, D / 2, ' Ronchi Ruling', ha='left', color='black')

    # screen
    plt.axvline(2 * f, color='blue', linewidth=2)
    plt.text(2 * f * 1.02, -D / 2, 'projection screen', ha='left', rotation=90, color='blue')

    plt.xlabel('Distance from focus (mm)')
    plt.ylabel('Height above optical axis (mm)')
    plt.title('Ronchi Ruling Lens Test')
    return fig, ax


def plot_mirror_layout(D, RoC, z_offset):
    """
    Plot the Ronchi Mirror Test Layout.

    Args:
        D: diameter of mirror or lens [mm]
        RoC: radius of curvature of mirror [mm]
        z_offset: axial offset of knife edge from true focus [mm]
    Returns:
        fig: matplotlib Figure object representing the plot
        ax: matplotlib Axes object representing the plot
    """
    fig, ax = plt.subplots(figsize=(10, 5))

    # initial height offset
    yo = D / 8

    # point source
    plt.text(0, yo, ' point source', ha='left', va='center', color='blue')

    # marginal rays with mirror at -RoC
    DD = 0.8 * D
    y_screen = -yo - (DD / 2 + yo) / RoC * D
    plt.plot([0, -RoC, 0, D], [yo, DD / 2, -yo, y_screen], color='blue', linewidth=1)
#    plt.plot([0, -RoC, 0], [yo, DD / 4, -yo], color='black', linewidth=1)
#    plt.plot([0, -RoC, 0], [yo, -DD / 4, -yo], color='black', linewidth=1)
    y_screen = -yo - (DD / 2 - yo) / RoC * D
    plt.plot([0, -RoC, 0, D], [yo, -DD / 2, -yo, -y_screen - 2 * yo], color='black', linewidth=1)

    # draw the mirror
    lenstest.lenstest.draw_mirror(D, RoC, -RoC)
    plt.text(-RoC, D / 2, 'mirror under test', ha='left', color='blue')

# focus plane
#    plt.text(0, D / 2, ' focus', ha='left')
#    plt.axvline(0, color='black', linewidth = 1)

    # optical axis
    plt.axhline(0, color='blue', linewidth=1)
#    plt.text(-RoC * 0.9, 0, 'optical axis ', ha='left', va='center', color='blue',
#             bbox={"facecolor": "white", "edgecolor":"white"})

    # Ronchi
#    plt.axvline(z_offset, color='black', linewidth = 0.5)
    plt.plot([z_offset, z_offset], [0, -D / 2], ls='--', lw=2, color='black')
    plt.text(z_offset, -D / 2, ' Ronchi Ruling', ha='right', color='black')

    # screen
    plt.axvline(D, color='blue', linewidth=2)
    plt.text(D * 1.02, -D / 2, 'projection screen', ha='left', rotation=90, color='blue')

    plt.xlabel('Distance from focus (mm)')
    plt.ylabel('Height above optical axis (mm)')
    plt.title('Ronchi Ruling Lens Test')
    return fig, ax



# pylint: disable=invalid-name
# pylint: disable=unused-variable
"""Test functionality of utility routines found in lenstest.py."""

import unittest
import numpy as np
import lenstest.lenstest


class TestLensTest(unittest.TestCase):
    """Evaluate all the functions."""

    def test_sagitta(self):
        """Ensure basic execution with different parameters."""
        inputs = [
            (100, 0, 0, 0),  # Test a flat surface
            (100, 1, 0, 0, 0),  # Test a sphere
            (100, -0.5, 0, 0, 0),  # Test a prolate spheroid
            (100, -1, 0, 0, 0),  # Test a paraboloid
            (100, -2, 0, 0, 0),  # Test a hyperboloid
            (100, 0, 5, 10, 0),  # Test the function with non-zero X and Y values
            (100, 0, 5, 10, 0.1),  # Test the function with a non-zero offset
            (100, 1, np.array([1, 2]), np.array([0, 3]), 0),  # Test with an array
        ]

        for x in inputs:
            lenstest.lenstest.sagitta(*x)

    def test_draw_circle(self):
        """Ensure basic execution with different parameters."""
        inputs = [
            (100, 0, 0, 'black'),  # Test a circle with default X0, Y0, and color values
            (100, 50, 50, 'red'),  # Test a circle with non-default X0, Y0, and color values
            (100, -50, -50, 'blue'),  # Test a circle with negative X0 and Y0 values
        ]

        for x in inputs:
            lenstest.lenstest.draw_circle(*x)

    def test_XY_test_points(self):
        """Ensure basic execution with different parameters."""
        inputs = [
            (100, 100000, True),  # Test the function with default values
            (100, 100000, False),  # Test the function with non-random points
            (100, 5, True),  # Test the function with a smaller number of points
        ]

        for x in inputs:
            lenstest.lenstest.XY_test_points(*x)

    def test_knife_polygon(self):
        """Ensure basic execution with different parameters."""
        inputs = [
            (10, 0, 0),  # Test the function with default values
            (10, np.pi / 2, 0),  # Test the function with a non-zero phi value
            (10, 0, 5),  # Test the function with a non-zero dx value
        ]

        for x in inputs:
            lenstest.lenstest.knife_polygon(*x)

    def test_circle_polygon(self):
        """Ensure basic execution with different parameters."""
        inputs = [
            (100, 0, 0),  # Test the function with default values
            (100, 50, 50),  # Test the function with non-default X0 and Y0 values
            (100, -50, -50),  # Test the function with negative X0 and Y0 values
        ]

        for x in inputs:
            lenstest.lenstest.circle_polygon(*x)

    def test_sagitta_values(self):
        """
        Test calculations against Table 1 in Benjamin & Rosenblum.

        --- "Radii of Curvature and Sagittal Depths of Conic Sections, "
        ICLC, Vol. 19, pp. 76-83, March/April 1992

        The eccentricity for the first column was changed from -0.45 to 0.504j
        this gives the values of sagitta at all heights.

        I believe that there was a typo for results[8, 3].  This is fixed.
        """
        e = np.array([0.504j, 0, 0.45, 1, 2])
        conic_constant = (-e * e).real

        # table from the paper
        results = np.array([[1.0, 0.016, 0.016, 0.016, 0.016, 0.016],
                            [2.0, 0.064, 0.064, 0.064, 0.064, 0.063],
                            [3.0, 0.146, 0.146, 0.145, 0.144, 0.140],
                            [4.0, 0.262, 0.261, 0.260, 0.256, 0.245],
                            [5.0, 0.414, 0.412, 0.409, 0.401, 0.374],
                            [6.0, 0.606, 0.600, 0.595, 0.577, 0.524],
                            [7.0, 0.842, 0.829, 0.820, 0.785, 0.693],
                            [8.0, 1.128, 1.104, 1.086, 1.026, 0.878],
                            [9.0, 1.472, 1.429, 1.400, 1.298, 1.076],
                            [10.0, 1.890, 1.813, 1.761, 1.603, 1.285],
                            [11.0, 2.403, 2.269, 2.183, 1.939, 1.504],
                            [12.0, 3.061, 2.816, 2.673, 2.308, 1.731]])

        # Sagittal Depths Given for a Spherical Surface Having a Radius of Curvature of 7.80 mm,
        RoC = 7.8
        y = (results[:, 0] / 2).flatten()
        x = np.zeros_like(y)

        for i, K in enumerate(conic_constant):
            sag = lenstest.lenstest.sagitta(RoC, K, x, y)
            print(sag)
            table_sag = (results[:, i + 1]).flatten()
            print(table_sag)
            np.testing.assert_allclose(sag, table_sag, atol=3e-3, rtol=3e-3)


# Run the test case
if __name__ == '__main__':
    unittest.main()



# pylint: disable=invalid-name
# pylint: disable=too-many-arguments

"""
Utility routines for lens testing.

Documentation and examples are available at <https://lenstest.readthedocs.io>
"""

import numpy as np
import matplotlib.pyplot as plt

__all__ = ('sagitta',
           'XY_test_points',
           'knife_polygon',
           'circle_polygon',
           'draw_circle',
           'draw_lens',
           'draw_mirror',
           'conic_string',
           )


def sagitta(RoC, conic, X, Y, A=0, D=0):
    """
    Calculate sagitta for conic surface at points X, Y.

    This assumes that the point source of light is located at the center
    of the mirror radius of curvature.

    The Ronchi grating is located at RoC + z_offset and oriented so lines
    are perpendicular to the x-axis

    The conic section is specified by conic::
        conic = ∞ for surface that is flat,
        conic > 0 for surface that is an oblate spheroid,
        conic = 0 for surface that is a sphere,
        0<conic<-1 for surface that is a prolate spheroid,
        conic = -1 for surface that is a paraboloid
        conic < -1 for surface that is a hyperboloid

    Args:
        RoC: radius of curvature of mirror [mm]
        conic: conic constant or Schwartzchild constant [-]
        X: x-value(s) of points [mm]
        Y: y-value(s) of points [mm]
        A: spherical aberration [1/mm³]
        D: defocus [1/mm]

    Returns:
        sagitta at each (x, y) point [mm]
    """
    # find the x value where the ray passes through the Ronchi ruling
    rho_sqr = X**2 + Y**2
    heights = rho_sqr / (RoC + np.sqrt(RoC * RoC - (1 + conic) * rho_sqr))
    heights += A * rho_sqr**2
    heights += D * rho_sqr
    return heights


def draw_circle(R, X0=0, Y0=0, color='black'):
    """
    Draw a circle.

    If the circle that is plotted should be round, then
    `plt.gca().set_aspect('equal')` so aspect ratio on both axes is equal.

    Args:
        R: radius of circle
        X0: x-value of center of circle
        Y0: y-value of center of circle
        color: color of the circle
    Returns:
        X, Y: arrays of test points
    """
    theta = np.linspace(0, 2 * np.pi, 100)
    plt.plot(X0 + R * np.sin(theta), Y0 + R * np.cos(theta), color=color)


def XY_test_points(D, N=100000, on_grid=False):
    """
    Generate test points for lens test.

    The generated points all fall within a circle of radius D.

    Args:
        D: diameter of mirror [mm]
        N: number of points to generate
        on_grid: if False generate points on a grid
    Returns:
        X, Y: arrays of test points
    """
    if on_grid:
        gridpts = np.linspace(-D / 2, D / 2, int(np.sqrt(N)))
        x_grid, y_grid = np.meshgrid(gridpts, gridpts)
        r_mask = x_grid**2 + y_grid**2 > D * D / 4
        X = np.ma.masked_where(r_mask, x_grid)
        Y = np.ma.masked_where(r_mask, y_grid)
    else:
        U1 = np.random.uniform(size=N)
        U2 = np.random.uniform(size=N)
        X = D / 2 * np.sqrt(U2) * np.cos(2 * np.pi * U1)
        Y = D / 2 * np.sqrt(U2) * np.sin(2 * np.pi * U1)

    return X, Y


def knife_polygon(s, phi, ds):
    """
    Create a polygon for a rotated knife edge.

    The polygon is a vertical 1:2 rectangle that is shifted and
    then rotated counter clockwise.

    When the shift is zero, then the edge of the knife edge is at the origin.
    Specifing a shift translates the knife edge across the origin.  Then the
    knife is rotated about the origin by an angle phi.

    Args:
        s: short side of the rectangle   [mm]
        phi: CCW rotation from vertical  [radians]
        ds: shift of knife edge          [mm]
    Returns:
        x, y: coordinates of knife polygon
    """
    # vertical s x 2s rectangle with center of edge shifted by ds
    xp = np.array([-s, -s, 0, 0, -s]) + ds
    yp = np.array([-s, s, s, -s, -s])

    # rotate rectangle CCW
    x = xp * np.cos(phi) - yp * np.sin(phi)
    y = xp * np.sin(phi) + yp * np.cos(phi)

    return x, y


def circle_polygon(R, X0=0, Y0=0):
    """
    Create a polygon for a circle.

    If you want to plot the circle then the aspect ratio for the axes should
    to be set to equal: `plt.gca().set_aspect('equal')` so that the circle is
    looks round.

    Args:
        R: radius of the circle
        X0: x-value of center of circle
        Y0: y-value of center of circle
    Returns:
        x, y: coordinates of polygon
    """
    theta = np.linspace(0, 2 * np.pi, 100)
    return X0 + R * np.sin(theta), Y0 + R * np.cos(theta)


def draw_lens(D, RoC, middle=0):
    """
    Draw and fill a biconvex lens shape.

    If the glass has an index of 1.5, then the radius-of-curvature
    will equal the focal length.

    The lens is drawn with its center at `center`

    Args:
        D: diameter of the lens
        RoC: x-value of center of circle
        middle: y-value of center of circle
    Returns:
        Nothing.  Use `plt.show()` to see result
    """
    theta_max = np.arctan2(D / 2, RoC)
    theta = np.linspace(-theta_max, theta_max, 11)
    x = -RoC * np.cos(theta)
    y = -RoC * np.sin(theta)
    dx = RoC + x
    dx -= dx[0]     # offset surface to center the lens on the vertex

    xx = np.concatenate((middle - dx, middle + dx))
    yy = np.concatenate((y, -y))

    plt.fill(xx, yy, color='gray', alpha=0.8)


def draw_mirror(D, RoC, vertex=0):
    """Draw one face of a mirror."""
    theta_max = np.arctan2(D / 2, RoC)
    theta = np.linspace(-theta_max, theta_max, 51)
    x = -RoC * np.cos(theta)
    y = -RoC * np.sin(theta)
    dx = RoC + x
    dx -= dx[0]     # offset surface to center the lens on the vertex

    xx_last = np.array([vertex - dx[0] - D / 10, vertex - dx[0] - D / 10, vertex - dx[-1]])
    yy_last = np.array([y[-1], y[0], y[0]])
    xx = np.concatenate((vertex + dx, xx_last))
    yy = np.concatenate((y, yy_last))

    plt.fill(xx, yy, color='gray', alpha=0.8)


def conic_string(conic):
    """Create string that describes a conic value."""
    if np.isinf(conic):
        return 'flat'

    if conic > 0:
        return "oblate spheroid"

    if conic == 0:
        return "sphere"

    if conic > -1:
        return "prolate spheroid"

    if conic == -1:
        return "paraboloid"

    return "hyperboloid"
